abt.fit <- function(x,y,
                    offset = NULL,
                    misc = NULL,
                    distribution = "gaussian",
                    w = NULL,
                    var.monotone = NULL,
                    n.trees = 1000,
                    interaction.depth = 1,
                    n.minobsinnode = ceiling(log10(nrow(data))+1),
                    shrinkage = 0.01,
                    bag.fraction = 0.5,
                    train.fraction = 1.0,
                    keep.data = TRUE,
                    verbose = TRUE,
                    var.names = NULL,
                    response.name = NULL,
                    rf.nvar = rf.nvar)
{
   cRows <- nrow(x)
   cCols <- ncol(x)
   if(is.null(var.names))
   {
      if(is.matrix(x))
      {
         var.names <- colnames(x)
      }
      else if(is.data.frame(x))
      {
         var.names <- names(x)
      }
      else
      {
         var.names <- paste("X",1:cCols,sep="")
      }
   }
   if(is.null(response.name))
   {
      response.name <- "y"
   }

   # check dataset size
   if(cRows*train.fraction*bag.fraction <= 2*n.minobsinnode+1)
   {
      stop("The dataset size is too small or subsampling rate is too large: cRows*train.fraction*bag.fraction <= n.minobsinnode")
   }

   if(nrow(x) != ifelse(class(y)=="Surv", nrow(y), length(y)))
   {
      stop("The number of rows in x does not equal the length of y.")
   }

   if(interaction.depth < 1)
   {
      stop("interaction.depth must be at least 1.")
   }

   if(length(w)==0) w <- rep(1, cRows)
   else if(any(w < 0)) stop("negative weights not allowed")
   w <- w*length(w)/sum(w) # normalize to N

   if(any(is.na(y))) stop("Missing values are not allowed in the response, ",
                          response.name)

   if(is.null(offset) || (offset==0))
   {
      offset <- NA
   }
   else if(length(offset) != length(y))
   {
      stop("The length of offset does not equal the length of y.")
   }
   Misc <- NA

   # setup variable types
   var.type <- rep(0,cCols)
   var.levels <- vector("list",cCols)
   for(i in 1:length(var.type))
   {
      if(all(is.na(x[,i])))
      {
         stop("variable ",i,": ",var.names[i]," has only missing values.")
      }
      if(is.ordered(x[,i]))
      {
         var.levels[[i]] <- levels(x[,i])
         x[,i] <- as.numeric(x[,i])-1
         var.type[i] <- 0
      }
      else if(is.factor(x[,i]))
      {
      if(length(levels(x[,i]))>1024)
            stop("abt does not currently handle categorical variables with more than 1024 levels. Variable ",i,": ",var.names[i]," has ",length(levels(x[,i]))," levels.")
         var.levels[[i]] <- levels(x[,i])
         x[,i] <- as.numeric(x[,i])-1
         var.type[i] <- max(x[,i],na.rm=TRUE)+1
      }
      else if(is.numeric(x[,i]))
      {
         var.levels[[i]] <- quantile(x[,i],prob=(0:10)/10,na.rm=TRUE)
      }
      else
      {
         stop("variable ",i,": ",var.names[i]," is not of type numeric, ordered, or factor.")
      }

      # check for some variation in each variable
      if(length(unique(var.levels[[i]])) == 1)
      {
         warning("variable ",i,": ",var.names[i]," has no variation.")
      }
   }

   nTrain <- as.integer(train.fraction*cRows)
   nClass <- 1

 # check potential problems with the distributions
   if((distribution$name == "bernoulli") && !all(0<=y & y<=1))
   {
      stop("Bernoulli requires the response to be in the range [0,1]")
   }
   if((distribution$name == "huberized") && !all(0<=y & y<=1))
   {
      stop("Huberized requires the response to be in the range [0,1]")
   }
   if((distribution$name == "poisson") && any(y<0))
   {
      stop("Poisson requires the response to be positive")
   }

   if((distribution$name == "adaboost") && !all(0<=y & y<=1))
   {
      stop("This version of AdaBoost requires the response to be in {0,1}")
   }
   if(distribution$name == "quantile")
   {
      if(length(unique(w)) > 1)
      {
         stop("This version of gbm for the quantile regression lacks a weighted quantile. For now the weights must be constant.")
      }
      if(is.null(distribution$alpha))
      {
         stop("For quantile regression, the distribution parameter must be a list with a parameter 'alpha' indicating the quantile, for example list(name=\"quantile\",alpha=0.95).")
      } else
      if((distribution$alpha<0) || (distribution$alpha>1))
      {
         stop("alpha must be between 0 and 1.")
      }
      Misc <- c(alpha=distribution$alpha)
   }
 
   if(distribution$name == "coxph")
   {  
      require(survival)  
      if(class(y)!="Surv")
      {
         stop("Outcome must be a survival object Surv(time,failure)")
      }
      if(attr(y,"type")!="right")
      {
         stop("abt() currently only handles right censored observations")
      }
      Misc <- y[,2]
      y <- y[,1]

      # reverse sort the failure times to compute risk sets on the fly
      i.train <- order(-y[1:nTrain])
      n.test <- cRows - nTrain
      if(n.test > 0)
      {
         i.test <- order(-y[(nTrain+1):cRows]) + nTrain
      }
      else
      {
         i.test <- NULL
      }
      i.timeorder <- c(i.train,i.test)

      y <- y[i.timeorder]
      Misc <- Misc[i.timeorder]
      x <- x[i.timeorder,,drop=FALSE]
      w <- w[i.timeorder]
      if(!is.na(offset)) offset <- offset[i.timeorder]
   }

   if(distribution$name == "bisquare")   {
          Misc <- estBisqParam( distribution$eff )
   }
   if(distribution$name == "tdist")
   {
        if (is.null(distribution$df) || !is.numeric(distribution$df)){
         Misc <- 4
        }
        else {
         Misc <- distribution$df[1]
        }
   }

  if (distribution$name == "multinomial")
   {
      ## Ensure that the training set contains all classes
      classes <- attr(factor(y), "levels")
      nClass <- length(classes)

      if (nClass > nTrain){
         stop(paste("Number of classes (", nClass, 
                    ") must be less than the size of the training set (", nTrain, ")", 
                    sep = ""))
      }

      new.idx <- as.vector(sapply(classes, function(a,x){ min((1:length(x))[x==a]) }, y))

      all.idx <- 1:length(y)
      new.idx <- c(new.idx, all.idx[!(all.idx %in% new.idx)])

      y <- y[new.idx]
      x <- x[new.idx, ,drop=F]
      w <- w[new.idx]
      if (!is.null(offset)){
         offset <- offset[new.idx]
      }
   
      ## Get the factors
      y <- as.numeric(as.vector(outer(y, classes, "==")))
      
      ## Fill out the weight and offset
      w <- rep(w, nClass)
      if (!is.null(offset)){
         offset <- rep(offset, nClass)
      }
   } 
   
   if(!is.numeric(y))
   {
      stop("The response ",response.name," must be numeric. Factors must be converted to numeric")
   }

   # create index upfront... subtract one for 0 based order
   x.order <- apply(x[1:nTrain,,drop=FALSE],2,order,na.last=FALSE)-1
 
   x <- as.vector(data.matrix(x))
   predF <- rep(0,length(y))
   train.error <- rep(0,n.trees)
   valid.error <- rep(0,n.trees)
   oobag.improve <- rep(0,n.trees)

   if(is.null(var.monotone)) var.monotone <- rep(0,cCols)
   else if(length(var.monotone)!=cCols)
   {
      stop("Length of var.monotone != number of predictors")
   }
   else if(!all(is.element(var.monotone,-1:1)))
   {
      stop("var.monotone must be -1, 0, or 1")
   }
   fError <- FALSE
 
   abt.obj <- .Call("gbm",
                    Y=as.double(y),
                    Offset=as.double(offset),
                    X=as.double(x),
                    X.order=as.integer(x.order),
                    weights=as.double(w),
                    Misc=as.double(Misc),
                    cRows=as.integer(cRows),
                    cCols=as.integer(cCols),
                    var.type=as.integer(var.type),
                    var.monotone=as.integer(var.monotone),
                    distribution=as.character(distribution$name),
                    rf.nvar=as.integer(rf.nvar),
                    n.trees=as.integer(n.trees),
                    interaction.depth=as.integer(interaction.depth),
                    n.minobsinnode=as.integer(n.minobsinnode),
                    n.classes = as.integer(nClass),
                    shrinkage=as.double(shrinkage),
                    bag.fraction=as.double(bag.fraction),
                    nTrain=as.integer(nTrain),
                    fit.old=as.double(NA),
                    n.cat.splits.old=as.integer(0),
                    n.trees.old=as.integer(0),
                    verbose=as.integer(verbose),
                    PACKAGE = "abt")
                    
   names(abt.obj) <- c("initF","fit","train.error","valid.error",
                       "oobag.improve","trees","c.splits")

   abt.obj$bag.fraction <- bag.fraction
   abt.obj$distribution <- distribution
   abt.obj$interaction.depth <- interaction.depth
   abt.obj$n.minobsinnode <- n.minobsinnode
   abt.obj$rf.nvar <- rf.nvar
   abt.obj$num.classes <- nClass
   abt.obj$n.trees <- length(abt.obj$trees) / nClass
   abt.obj$nTrain <- nTrain
   abt.obj$response.name <- response.name
   abt.obj$shrinkage <- shrinkage
   abt.obj$train.fraction <- train.fraction
   abt.obj$var.levels <- var.levels
   abt.obj$var.monotone <- var.monotone
   abt.obj$var.names <- var.names
   abt.obj$var.type <- var.type
   abt.obj$verbose <- verbose
   abt.obj$Terms <- NULL

   if(distribution$name == "coxph")
   {
      abt.obj$fit[i.timeorder] <- abt.obj$fit
   }

   ## If K-Classification is used then split the fit and tree components
   if (distribution$name == "multinomial"){
      abt.obj$fit <- matrix(abt.obj$fit, ncol = nClass)
      dimnames(abt.obj$fit)[[2]] <- classes
      abt.obj$classes <- classes
          
      ## Also get the class estimators
      exp.f <- exp(abt.obj$fit)
      denom <- matrix(rep(rowSums(exp.f), nClass), ncol = nClass)
      abt.obj$estimator <- exp.f/denom
   }

   if(keep.data)
   {
      if(distribution$name == "coxph")
      {
         # put the observations back in order
         abt.obj$data <- list(y=y,x=x,x.order=x.order,offset=offset,Misc=Misc,w=w,
                              i.timeorder=i.timeorder)
      }
      else if ( distribution$name == "multinomial" ){
         # Restore original order of the data
         new.idx <- order( new.idx )
         abt.obj$data <- 
         	list( y=as.vector(matrix(y, ncol=length(classes),byrow=FALSE)[new.idx,]),
            x=as.vector(matrix(x, ncol=length(var.names), byrow=FALSE)[new.idx,]),
            x.order=x.order, offset=offset[new.idx], Misc=Misc, w=w[new.idx] )
      }
      else
      {
         abt.obj$data <- list(y=y,x=x,x.order=x.order,offset=offset,Misc=Misc,w=w)
      }
   }
   else
   {
      abt.obj$data <- NULL
   }

   class(abt.obj) <- "abt"
   return(abt.obj)
}

