abtMC <- function(formula = formula(data), data = list(), 
    distribution = "gaussian", weights, offset = NULL, var.monotone = NULL, 
    n.trees = 1000, interaction.depth = 3, n.minobsinnode = round(log10(nrow(data))), 
    shrinkage = 0.01, bag.fraction = 0.5, keep.data = FALSE,
    verbose = FALSE, monitor = TRUE, use = c("best", "all")[1], 
    prune = TRUE, na.omit.y = TRUE, cv.folds = 5, gcv = 0, gncv = FALSE, 
    single.tree = FALSE, use.OOB = 1.5, train.fraction = 1, rf = FALSE, 
    rf.nvar = 0, class.stratify.cv = TRUE, stratify = TRUE, seed.stratify = 0, 
    seed.bag = 0) {
    
	require(foreach)
	require(doMC)	
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "weights", "offset"), names(mf), 
        0)
    mf <- mf[c(1, m)]
    mf$drop.unused.levels <- TRUE
    mf$na.action <- na.pass
    mf[[1]] <- as.name("model.frame")
    mf <- eval(mf, parent.frame())
    Terms <- attr(mf, "terms")
    
    y <- model.response(mf)
    w <- model.weights(mf)
    offset <- model.offset(mf)
    
    var.names <- attributes(Terms)$term.labels
    x <- model.frame(terms(reformulate(var.names)), data, na.action = na.pass)
    
    # get the character name of the response variable
    response.name <- as.character(formula[[2]])
    
    if (na.omit.y & any(nays <- is.na(y))) {
        if (verbose) 
            cat("Missing responses; ", sum(nays), "cases deleted\n")
        y <- y[!nays]
        x <- x[!nays, ]
        w <- w[!nays]
        if (!is.null(offset)) 
            offset <- offset[!nays]
    }
    
    nrows <- nrow(x)
    ncols <- ncol(x)
    
    if (is.character(distribution)) 
        distribution <- list(name = distribution)
    if (!("name" %in% names(distribution))) {
        stop("The distribution is missing a 'name' component, for example list(name=\"gaussian\")")
    }
    supported.distributions <- c("bernoulli", "gaussian", "poisson", 
        "adaboost", "laplace", "coxph", "quantile", "bisquare", 
        "tdist", "multinomial", "huberized", "locationm")
    distribution$name <- supported.distributions[pmatch(distribution$name, 
        supported.distributions)]
    if (is.na(distribution$name)) {
        stop("Distribution ", distribution$name, " is not supported")
    }
    
    btype <- c("best", "ibest", "all")
    use <- btype[pmatch(use, btype)]
    
    if (rf & rf.nvar == 0) 
        rf.nvar <- floor(sqrt(ncols))
    
    if (monitor) {
        cat(" N cases =", nrows, ": N trees =", n.trees, ": Depth =", 
            interaction.depth, ": Minimum node size =", n.minobsinnode, 
            "\n")
        cat(" Distribution =", distribution$name, ": Shrinkage =", 
            signif(shrinkage), ": Bag Fraction =", bag.fraction, 
            "\n")
        if (rf) 
            cat(" Random inputs using", rf.nvar, "from", ncols, 
                "\n")
    }
    
    if (cv.folds > 0) {
        lst <- list()
        nontr <- list()
        if (seed.stratify > 0) 
            set.seed(seed.stratify)
        if (gcv == 0) {
            if (!stratify) 
                nsamps <- sample(rep(1:cv.folds, length = nrows))
            else nsamps <- rep(sample(1:cv.folds), ceiling(nrows/cv.folds))[1:nrows][order(y)]
        }
        else {
            if (na.omit.y & any(nays)) {
                if (gncv) 
                  cv.folds <- length(unique(data[nays, gcv]))
                nsamps <- cv.var(data[!nays, gcv], cv.folds)
            }
            else {
                if (gncv) 
                  cv.folds <- length(unique(data[, gcv]))
                nsamps <- cv.var(data[, gcv], cv.folds)
            }
            if ((luni <- length(unique(nsamps))) < cv.folds) 
                cv.folds <- luni
        }
        if (seed.bag > 0) 
            set.seed(seed.bag)
		lst <- foreach(i = 1:cv.folds) %do% {
            nTrain <- length(trainees <- (1:nrows)[nsamps != 
                i])
            nontr[[i]] <- non.trainees <- (1:nrows)[is.na(match(1:nrows, 
                trainees))]
            temp.x <- x[c(trainees, non.trainees), , drop = F]
            temp.y <- y[c(trainees, non.trainees)]
            temp.w <- if (is.null(w)) 
                logical(0)
            else w[c(trainees, non.trainees)]
            temp.offset <- offset[c(trainees, non.trainees)]
            lst0 <- abt.fit(temp.x, temp.y, offset = temp.offset, 
                distribution = distribution, w = temp.w, var.monotone = var.monotone, 
                n.trees = n.trees, interaction.depth = interaction.depth, 
                n.minobsinnode = n.minobsinnode, shrinkage = shrinkage, 
                bag.fraction = bag.fraction, train.fraction = nTrain/nrows, 
                keep.data = keep.data, verbose = verbose, var.names = var.names, 
                response.name = response.name, rf.nvar = rf.nvar)
            lst0$Terms <- Terms
        	lst0
        }
        n.trees <- lst[[1]]$n.trees
        cv.error <- rep(0, n.trees)
        for (i in 1:cv.folds) cv.error <- cv.error + lst[[i]]$valid.error * 
            (nrows - lst[[i]]$nTrain)
        cv.error <- cv.error/nrows
        if (use == "best") {
            lst[[1]]$best <- best <- which.min(cv.error)
            lst[[1]]$best.error <- best.error <- cv.error[best]
            lst[[1]]$cv.error <- cv.error
            if (prune) {
                for (i in 1:cv.folds) {
                  lst[[i]]$trees <- lst[[i]]$trees[1:best]
                  lst[[i]]$valid.error <- lst[[i]]$valid.error[1:best]
                  lst[[i]]$train.error <- lst[[i]]$train.error[1:best]
                  lst[[i]]$oobag.improve <- lst[[i]]$oobag.improve[1:best]
                  lst[[i]]$n.trees <- lst[[i]]$best
                }
            }
        }
        else {
            lst[[1]]$best <- best <- n.trees
            lst[[1]]$best.error <- best.error <- cv.error[best]
            lst[[1]]$cv.error <- cv.error
        }
        
        if (distribution$name == "multinomial") 
            fitted <- matrix(NA, nrow = nrows, ncol = length(unique(y)))
        else fitted <- rep(NA, length = nrows)
        
        for (i in 1:cv.folds) if (distribution$name != "multinomial") 
            fitted[nontr[[i]]] <- predict(lst[[i]], newdata = x[nontr[[i]], 
                , drop = F], n.trees = ifelse(use == "best", 
                best, n.trees))
        else fitted[nontr[[i]], ] <- predict(lst[[i]], newdata = x[nontr[[i]], 
            , drop = F], n.trees = ifelse(use == "best", best, 
            n.trees))
        lst[[1]]$rf.nvar <- rf.nvar
        lst[[1]]$mdata <- mf[!nays, ]
        lst[[1]]$nsamps <- nsamps
        lst[[1]]$fitted <- fitted
        lst[[1]]$residuals <- y - fitted
        lst[[1]]$m <- mf
        names(lst) <- paste("abt", 1:cv.folds, sep = "-")
        class(lst) <- c("abt", "multi")
    }
    if (single.tree || cv.folds == 0) {
        if (cv.folds > 0) {
            n.trees <- best
        }
        lst <- abt.fit(x = x, y = y, offset = offset, distribution = distribution, 
            w = w, var.monotone = var.monotone, n.trees = n.trees, 
            interaction.depth = interaction.depth, n.minobsinnode = n.minobsinnode, 
            shrinkage = shrinkage, bag.fraction = bag.fraction, 
            train.fraction = train.fraction, keep.data = keep.data, 
            verbose = verbose, var.names = var.names, response.name = response.name, 
            rf.nvar = rf.nvar)
        lst$Terms <- Terms
        if (cv.folds > 1) {
            lst$cv.error <- cv.error
            lst$best.error <- best.error
            lst$best <- n.trees
        }
        else if (use.OOB) {
            lst$best <- round(use.OOB * abt.perf(lst, meth = "OOB", 
                plot = F))
        }
        lst$mdata <- mf[!nays, ]
        lst$fitted <- fitted
        lst$residuals <- y - fitted
        lst$m <- mf
        class(lst) <- c("abt", "single")
    }
    if (monitor) {
        if (cv.folds) 
            cat(" CV-folds =", cv.folds, " :")
        if (use.OOB) 
            cat(" Best size =", best, ": Best error =", signif(best.error, 
                4), "\n")
        else cat("\n")
    }
    return(lst)
}

